import type { ActionGroupSet } from 'api-state-handlers';
import type { RecordRef, FileSet, JsonData } from 'toolkit-data-structures';
import type {
  AuthInstructions,
  FlistInstructions,
  PlistInstructions,
  SlistInstructions,
  RemoveInstructions,
  RetrieveInstructions,
  StoreInstructions,
  TaskInstructions,
  UpdateInstructions,
} from 'api-state-instructions';
import type { Constructor } from 'toolkit-data-structures';
import type { RecordConfig } from 'toolkit-config';
import type { PlistQuery, SlistQuery, FlistQuery } from 'toolkit-list-structures';

import {
  StoreHandler,
  UpdateHandler,
  RetrieveHandler,
  FlistHandler,
  PlistHandler,
  SlistHandler,
  RemoveHandler,
  AuthHandler,
  TaskHandler,
} from 'api-state-handlers';

import { getQueryHandler } from 'records-query-handler';

import { RecordHandler, AuthImplementation } from 'records-handler-unit';

import { GetFlagsContract } from 'toolkit-contracts';

//

interface ActionGroup {
  // domain -> repository -> method = handler
  stores: { [key: string]: { [key: string]: Constructor<StoreHandler, [string]> } };
  updates: { [key: string]: { [key: string]: Constructor<UpdateHandler, [string]> } };
  retrievers: { [key: string]: { [key: string]: Constructor<RetrieveHandler, [string]> } };
  flists: { [key: string]: { [key: string]: Constructor<FlistHandler, [string]> } }; // todo here: flist handler
  plists: { [key: string]: { [key: string]: Constructor<PlistHandler, [string]> } }; // todo here: plist handler
  slists: { [key: string]: { [key: string]: Constructor<SlistHandler, [string]> } }; // todo here: slist handler
  removers: { [key: string]: { [key: string]: Constructor<RemoveHandler, [string]> } };
  auths: { [key: string]: { [key: string]: Constructor<AuthHandler, [string]> } };
  tasks: { [key: string]: { [key: string]: Constructor<TaskHandler, [string]> } };
  configs: { [key: string]: RecordConfig };
}

//

const camelToKebab = (str: string): string => {
  str = str.charAt(0).toLowerCase() + str.substr(1);
  return str.replace(/([a-z0-9]|(?=[A-Z]))([A-Z])/g, '$1-$2').toLowerCase();
};

//

const getActionGroups = (actionGroups: ActionGroupSet): ActionGroup => {
  const stores: { [key: string]: { [key: string]: Constructor<StoreHandler, [string]> } } = {};
  const updates: { [key: string]: { [key: string]: Constructor<UpdateHandler, [string]> } } = {};
  const retrievers: { [key: string]: { [key: string]: Constructor<RetrieveHandler, [string]> } } = {};
  const flists: { [key: string]: { [key: string]: Constructor<FlistHandler, [string]> } } = {};
  const plists: { [key: string]: { [key: string]: Constructor<PlistHandler, [string]> } } = {};
  const slists: { [key: string]: { [key: string]: Constructor<SlistHandler, [string]> } } = {};
  const removers: { [key: string]: { [key: string]: Constructor<RemoveHandler, [string]> } } = {};
  const auths: { [key: string]: { [key: string]: Constructor<AuthHandler, [string]> } } = {};
  const tasks: { [key: string]: { [key: string]: Constructor<TaskHandler, [string]> } } = {};
  const configs: { [key: string]: RecordConfig } = {};

  for (let [repository, actionGroup] of Object.entries(actionGroups)) {
    // stores
    if (!stores[repository]) {
      stores[repository] = {};
    }
    for (let creator of actionGroup.stores) {
      stores[repository][camelToKebab(creator.prototype.constructor.name)] = creator;
    }

    // updates
    if (!updates[repository]) {
      updates[repository] = {};
    }
    for (let updater of actionGroup.updates) {
      updates[repository][camelToKebab(updater.prototype.constructor.name)] = updater;
    }

    // retrievers
    if (!retrievers[repository]) {
      retrievers[repository] = {};
    }
    for (let retriever of actionGroup.retrievers) {
      retrievers[repository][camelToKebab(retriever.prototype.constructor.name)] = retriever;
    }

    // flists
    if (!flists[repository]) {
      flists[repository] = {};
    }
    for (let flist of actionGroup.flists) {
      flists[repository][camelToKebab(flist.prototype.constructor.name)] = flist;
    }

    // plists
    if (!plists[repository]) {
      plists[repository] = {};
    }
    for (let plist of actionGroup.plists) {
      plists[repository][camelToKebab(plist.prototype.constructor.name)] = plist;
    }

    // slists
    if (!slists[repository]) {
      slists[repository] = {};
    }
    for (let slist of actionGroup.slists) {
      slists[repository][camelToKebab(slist.prototype.constructor.name)] = slist;
    }

    // removers
    if (!removers[repository]) {
      removers[repository] = {};
    }
    for (let remover of actionGroup.removers) {
      removers[repository][camelToKebab(remover.prototype.constructor.name)] = remover;
    }

    // auths
    if (!auths[repository]) {
      auths[repository] = {}; //
    }
    for (let auth of actionGroup.auths) {
      auths[repository][camelToKebab(auth.prototype.constructor.name)] = auth;
    }

    // tasks
    if (!tasks[repository]) {
      tasks[repository] = {};
    }
    for (let task of actionGroup.tasks) {
      tasks[repository][camelToKebab(task.prototype.constructor.name)] = task;
    }

    // config
    configs[repository] = actionGroup.config;
  }

  return {
    stores,
    updates,
    retrievers,
    flists,
    plists,
    slists,
    removers,
    auths,
    tasks,
    configs,
  };
};

//

const getStoreInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  payload: JsonData,
  files: FileSet,
  recordConfig: RecordConfig,
): Promise<StoreInstructions> => {
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    filter: null,
    fill: null,
    authRecord,
    files,
    target: null,
    targetRef,
    payload,
    storeResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getUpdateInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  payload: JsonData,
  files: FileSet,
  recordConfig: RecordConfig,
): Promise<UpdateInstructions> => {
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    filter: null,
    fill: null,
    authRecord,
    files,
    target: null,
    targetRef,
    payload,
    updateResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getRetrieveInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  recordConfig: RecordConfig,
): Promise<RetrieveInstructions> => {
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    authRecord,
    target: null,
    targetRef,
    retrieveResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getFlistInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  flistQuery: FlistQuery,
  recordConfig: RecordConfig,
): Promise<FlistInstructions> => {
  const query = getQueryHandler();
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  flistQuery.builder = query(flistQuery.repository);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    authRecord,
    target: null,
    targetRef,
    flistQuery,
    listResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getPlistInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  plistQuery: PlistQuery,
  recordConfig: RecordConfig,
): Promise<PlistInstructions> => {
  const query = getQueryHandler();
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  plistQuery.builder = query(plistQuery.repository);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    authRecord,
    target: null,
    targetRef,
    plistQuery,
    listResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getSlistInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  slistQuery: SlistQuery,
  recordConfig: RecordConfig,
): Promise<SlistInstructions> => {
  const query = getQueryHandler();
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  slistQuery.builder = query(slistQuery.repository);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    authRecord,
    target: null,
    targetRef,
    slistQuery,
    listResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getRemoveInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  recordConfig: RecordConfig,
): Promise<RemoveInstructions> => {
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    authRecord,
    target: null,
    targetRef,
    removeResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getAuthInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  payload: JsonData,
  files: FileSet,
  recordConfig: RecordConfig,
): Promise<AuthInstructions> => {
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  return {
    recordConfig,
    recordHandler,
    filter: null,
    fill: null,
    authRecord,
    target: null,
    targetRef,
    payload,
    files,
    authResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

const getTaskInstructions = async (
  authImplementation: AuthImplementation,
  authorization: string | null,
  targetRef: RecordRef,
  payload: JsonData,
  files: FileSet,
  recordConfig: RecordConfig,
): Promise<TaskInstructions> => {
  const recordHandler = new RecordHandler(authImplementation);
  const authRecord = await recordHandler.auth().token(authorization);
  return {
    recordConfig,
    recordHandler,
    requiredAuthRecords: null,
    filter: null,
    fill: null,
    authRecord,
    target: null,
    targetRef,
    payload,
    files,
    taskResult: null,
    result: null,
    resultMsg: null,
    resultData: {},
    success: false,
    sideEffect: null,
    getFlags: GetFlagsContract,
  };
};

//

export {
  getActionGroups,
  getStoreInstructions,
  getUpdateInstructions,
  getRetrieveInstructions,
  getFlistInstructions,
  getPlistInstructions,
  getSlistInstructions,
  getRemoveInstructions,
  getAuthInstructions,
  getTaskInstructions,
};
